@@ -15,30 +15,22 @@
 #endif
 
 int bookmarks_codepage=0;
-int can_write_bookmarks=0;	/* global flag if we can write bookmarks */
 
 unsigned char bookmarks_file[MAX_STR_LEN]="";
-void *bookmark_new_item(void *);
-unsigned char *bookmark_type_item(struct terminal *, void *, int);
-void bookmark_delete_item(void *);
-void bookmark_edit_item(struct dialog_data *,void *,void (*)(struct dialog_data *,void *,void *,struct list_description *),void *, unsigned char);
-void bookmark_copy_item(void *, void *);
-void bookmark_goto_item(struct session *, void *);
-void *bookmark_default_value(struct session*, unsigned char);
-void *bookmark_find_item(void *start, unsigned char *str, int direction);
-void free_bookmarks(void);
-void bookmark_edit_item_fn(struct dialog_data *);
-void bookmark_edit_done(void *);
-void bookmark_edit_abort(struct dialog_data *);
-struct bookmark_list *previous_on_this_level(struct bookmark_list *);
-void add_bookmark(unsigned char *, unsigned char *, int);
-void create_initial_bookmarks(void);
-void load_bookmarks(void);
 
+static void *bookmark_new_item(void *);
+static unsigned char *bookmark_type_item(struct terminal *, void *, int);
+static void bookmark_delete_item(void *);
+static void bookmark_edit_item(struct dialog_data *,void *,void (*)(struct dialog_data *,void *,void *,struct list_description *),void *, unsigned char);
+static void bookmark_copy_item(void *, void *);
+static void bookmark_goto_item(struct session *, void *);
+static void *bookmark_default_value(struct session*, unsigned char);
+static void *bookmark_find_item(void *start, unsigned char *str, int direction);
+static void save_bookmarks(struct session *ses);
 
 struct list bookmarks={&bookmarks,&bookmarks,0,-1,NULL};
 
-struct history bookmark_search_history = { 0, { &bookmark_search_history.items, &bookmark_search_history.items } };
+static struct history bookmark_search_history = { 0, { &bookmark_search_history.items, &bookmark_search_history.items } };
 
 /* when you change anything, don't forget to change it in reinit_bookmarks too !*/
 
@@ -63,7 +55,7 @@ struct bookmark_list{
 };
 
 
-struct list_description bookmark_ld=
+static struct list_description bookmark_ld=
 {
 	1,  /* 0= flat; 1=tree */
 	&bookmarks,  /* list */
@@ -76,7 +68,6 @@ struct list_description bookmark_ld=
 	bookmark_find_item,
 	&bookmark_search_history,
 	0,		/* this is set in init_bookmarks function */
-	60,  /* width of main window */
 	15,  /* # of items in main window */
 	T_BOOKMARK,
 	T_BOOKMARKS_ALREADY_IN_USE,
@@ -84,8 +75,9 @@ struct list_description bookmark_ld=
 	T_DELETE_BOOKMARK,
 	T_GOTO,
 	bookmark_goto_item,	/* FIXME: should work (URL in UTF8), but who knows? */
+	save_bookmarks,
 
-	0,0,0,0,  /* internal vars */
+	NULL,NULL,0,0,  /* internal vars */
 	0, /* modified */
 	NULL,
 	NULL,
@@ -101,7 +93,7 @@ struct kawasaki
 
 
 /* clears the bookmark list */
-void free_bookmarks(void) 
+static void free_bookmarks(void) 
 {
 	struct bookmark_list *bm;
 
@@ -118,7 +110,6 @@ void free_bookmarks(void)
 /* called before exiting the links */
 void finalize_bookmarks(void)
 {
-	save_bookmarks();
 	free_bookmarks();
 }
 
@@ -127,7 +118,7 @@ void finalize_bookmarks(void)
 /* allocates struct kawasaki and puts current page title and url */
 /* type: 0=item, 1=directory */
 /* on error returns NULL */
-void *bookmark_default_value(struct session *ses, unsigned char type)
+static void *bookmark_default_value(struct session *ses, unsigned char type)
 {
 	struct kawasaki *zelena;
 	unsigned char *txt;
@@ -166,7 +157,7 @@ void *bookmark_default_value(struct session *ses, unsigned char type)
 }
 
 
-void bookmark_copy_item(void *in, void *out)
+static void bookmark_copy_item(void *in, void *out)
 {
 	struct bookmark_list *item_in=(struct bookmark_list*)in;
 	struct bookmark_list *item_out=(struct bookmark_list*)out;
@@ -190,14 +181,14 @@ void bookmark_copy_item(void *in, void *out)
 }
 
 
-unsigned char *bm_add_msg[] = {
-	TEXT(T_NNAME),
-	TEXT(T_URL),
+static unsigned char *bm_add_msg[] = {
+	TEXT_(T_NNAME),
+	TEXT_(T_URL),
 };
 
 
 /* Called to setup the add bookmark dialog */
-void bookmark_edit_item_fn(struct dialog_data *dlg)
+static void bookmark_edit_item_fn(struct dialog_data *dlg)
 {
 	int max = 0, min = 0;
 	int w, rw;
@@ -250,7 +241,7 @@ void bookmark_edit_item_fn(struct dialog_data *dlg)
 
 
 /* Puts url and title into the bookmark item */
-void bookmark_edit_done(void *data)
+static void bookmark_edit_done(void *data)
 {
 	struct dialog *d=(struct dialog*)data;
 	struct bookmark_list *item=(struct bookmark_list *)d->udata;
@@ -289,7 +280,7 @@ void bookmark_edit_done(void *data)
 
 
 /* destroys an item, this function is called when edit window is aborted */
-void bookmark_edit_abort(struct dialog_data *data)
+static void bookmark_edit_abort(struct dialog_data *data)
 {
 	struct bookmark_list *item=(struct bookmark_list*)data->dlg->udata;
 	struct dialog *dlg=data->dlg;
@@ -301,7 +292,7 @@ void bookmark_edit_abort(struct dialog_data *data)
 
 /* dlg_title is TITLE_EDIT or TITLE_ADD */
 /* edit item function */
-void bookmark_edit_item(struct dialog_data *dlg,void *data,void (*ok_fn)(struct dialog_data *,void * ,void *, struct list_description *),void * ok_arg, unsigned char dlg_title)
+static void bookmark_edit_item(struct dialog_data *dlg,void *data,void (*ok_fn)(struct dialog_data *,void * ,void *, struct list_description *),void * ok_arg, unsigned char dlg_title)
 {
 	struct bookmark_list *item=(struct bookmark_list *)data;
 	unsigned char *title, *url;
@@ -343,13 +334,13 @@ void bookmark_edit_item(struct dialog_data *dlg,void *data,void (*ok_fn)(struct
 	switch (dlg_title)
 	{
 		case TITLE_EDIT:
-		if ((item->type)&1)d->title=TEXT(T_EDIT_FOLDER);
-		else d->title=TEXT(T_EDIT_BOOKMARK);
+		if ((item->type)&1)d->title=TEXT_(T_EDIT_FOLDER);
+		else d->title=TEXT_(T_EDIT_BOOKMARK);
 		break;
 
 		case TITLE_ADD:
-		if ((item->type)&1)d->title=TEXT(T_ADD_FOLDER);
-		else d->title=TEXT(T_ADD_BOOKMARK);
+		if ((item->type)&1)d->title=TEXT_(T_ADD_FOLDER);
+		else d->title=TEXT_(T_ADD_BOOKMARK);
 		break;
 
 		default:
@@ -380,11 +371,11 @@ void bookmark_edit_item(struct dialog_data *dlg,void *data,void (*ok_fn)(struct
 	d->items[a+1].type = D_BUTTON;
 	d->items[a+1].gid = B_ENTER;
 	d->items[a+1].fn = ok_dialog;
-	d->items[a+1].text = TEXT(T_OK);
+	d->items[a+1].text = TEXT_(T_OK);
 	
 	d->items[a+2].type = D_BUTTON;
 	d->items[a+2].gid = B_ESC;
-	d->items[a+2].text = TEXT(T_CANCEL);
+	d->items[a+2].text = TEXT_(T_CANCEL);
 	d->items[a+2].fn = cancel_dialog;
 	
 	d->items[a+3].type = D_END;
@@ -395,7 +386,7 @@ void bookmark_edit_item(struct dialog_data *dlg,void *data,void (*ok_fn)(struct
 
 /* create new bookmark item and returns pointer to it, on error returns 0*/
 /* bookmark is filled with given data, data are deallocated afterwards */
-void *bookmark_new_item(void * data)
+static void *bookmark_new_item(void * data)
 {
 	struct bookmark_list *b;
 	struct kawasaki *zelena=(struct kawasaki *)data;
@@ -429,14 +420,14 @@ void *bookmark_new_item(void * data)
 
 /* allocate string and print bookmark into it */
 /* x: 0=type all, 1=type title only */
-unsigned char *bookmark_type_item(struct terminal *term, void *data, int x)
+static unsigned char *bookmark_type_item(struct terminal *term, void *data, int x)
 {
 	unsigned char *txt, *txt1;
 	struct bookmark_list* item=(struct bookmark_list*)data;
 	struct conv_table *table;
 
 	if (item==(struct bookmark_list*)(void *)&bookmarks)   /* head */
-		return stracpy(_(TEXT(T_BOOKMARKS),term));
+		return stracpy(_(TEXT_(T_BOOKMARKS),term));
 
 	txt=stracpy(item->title);
 #ifdef SHOW_URL
@@ -458,7 +449,7 @@ unsigned char *bookmark_type_item(struct terminal *term, void *data, int x)
 
 
 /* goto bookmark (called when goto button is pressed) */
-void bookmark_goto_item(struct session *ses, void *i)
+static void bookmark_goto_item(struct session *ses, void *i)
 {
 	struct bookmark_list *item=(struct bookmark_list*)i;
 
@@ -467,7 +458,7 @@ void bookmark_goto_item(struct session *ses, void *i)
 
 
 /* delete bookmark from list */
-void bookmark_delete_item(void *data)
+static void bookmark_delete_item(void *data)
 {
 	struct bookmark_list* item=(struct bookmark_list*)data;
 	struct bookmark_list *prev=item->prev;
@@ -492,7 +483,7 @@ static int substr_utf8(unsigned char *string, unsigned char *substr)
 	return r;
 }
 
-void * bookmark_find_item(void *start, unsigned char *str, int direction)
+static void * bookmark_find_item(void *start, unsigned char *str, int direction)
 {
 	struct bookmark_list *b,*s=(struct bookmark_list *)start;
 	
@@ -530,11 +521,12 @@ void * bookmark_find_item(void *start, unsigned char *str, int direction)
 
 /* returns previous item in the same folder and with same the depth, or father if there's no previous item */
 /* we suppose that previous items have correct pointer fotr */
-struct bookmark_list *previous_on_this_level(struct bookmark_list *item)
+static struct bookmark_list *previous_on_this_level(struct bookmark_list *item)
 {
 	struct bookmark_list *p;
 
-	for (p=item->prev;p->depth>item->depth;p=p->fotr);
+	for (p=item->prev;p->depth>item->depth;p=p->fotr)
+		;
 	return p;
 }
 
@@ -542,7 +534,7 @@ struct bookmark_list *previous_on_this_level(struct bookmark_list *item)
 /* create new bookmark at the end of the list */
 /* if url is NULL, create folder */
 /* both strings are null terminated */
-void add_bookmark(unsigned char *title, unsigned char *url, int depth)
+static void add_bookmark(unsigned char *title, unsigned char *url, int depth)
 {
 	struct bookmark_list *b,*p;
 	struct document_options *dop;
@@ -597,7 +589,7 @@ void add_bookmark(unsigned char *title, unsigned char *url, int depth)
 }
 
 /* Created pre-cooked bookmarks */
-void create_initial_bookmarks(void)
+static void create_initial_bookmarks(void)
 {
 	bookmarks_codepage=get_cp_index("8859-2");
 	add_bookmark("Links",NULL,0);
@@ -605,13 +597,13 @@ void create_initial_bookmarks(void)
 	add_bookmark("Calibration Procedure","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/calibration.html",2);
 	add_bookmark("Links Homepage","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/",2);
 	add_bookmark("Links Manual","http://links.twibright.com/user_en.html",2);
-	add_bookmark("esky",NULL,1);
-	add_bookmark("Kalibran procedura","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/kalibrace.html",2);
-	add_bookmark("Links: domc strnka","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/index_cz.html",2);
-	add_bookmark("Manul k Linksu","http://links.twibright.com/user.html",2);
+	add_bookmark("Cesky",NULL,1);
+	add_bookmark("Kalibracni procedura","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/kalibrace.html",2);
+	add_bookmark("Links: domaci stranka","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/index_cz.html",2);
+	add_bookmark("Manual k Linksu","http://links.twibright.com/user.html",2);
 }
 
-void load_bookmarks(void)
+static void load_bookmarks(struct session *ses)
 {
 	unsigned char *buf;
 	long len;
@@ -639,11 +631,10 @@ void load_bookmarks(void)
 	 */
 
 	buf=read_config_file(bookmarks_file);
-	can_write_bookmarks=1;
 	if (!buf){
-		can_write_bookmarks=(errno==ENOENT); /* if open failed and the file exists, don't write initial bookmarks when finishing */
 		create_initial_bookmarks();
 		bookmark_ld.modified=1;
+		save_bookmarks(ses);
 		return;
 	}
 
@@ -723,19 +714,45 @@ void init_bookmarks(void)
 		snprintf(bookmarks_file,MAX_STR_LEN,"%sbookmarks.html",links_home?links_home:(unsigned char*)"");
 
 	bookmark_ld.codepage=get_cp_index("utf-8");
-	load_bookmarks();
+	load_bookmarks(NULL);
 }
 
-void reinit_bookmarks(void)
+void reinit_bookmarks(struct session *ses, unsigned char *new_bookmarks_file, int new_bookmarks_codepage)
 {
-	free_bookmarks();
-	bookmarks.next=&bookmarks;
-	bookmarks.prev=&bookmarks;
-	bookmarks.type=0;
-	bookmarks.depth=-1;
-	bookmarks.fotr=NULL;
-	load_bookmarks();
-	reinit_list_window(&bookmark_ld);
+	unsigned char *buf;
+	if (bookmark_ld.open) {
+		msg_box(
+			ses->term,
+			NULL,
+			TEXT_(T_INFO),
+			AL_CENTER,
+			TEXT_(T_BOOKMARKS_ALREADY_IN_USE),
+			NULL,
+			1,
+			TEXT_(T_CANCEL),NULL,B_ENTER|B_ESC
+		);
+		return;
+	}
+
+	if (!strcmp(bookmarks_file, new_bookmarks_file)) {
+		goto save_only;
+	}
+
+	buf=read_config_file(new_bookmarks_file);
+	if (buf) {
+		mem_free(buf);
+		free_bookmarks();
+		safe_strncpy(bookmarks_file,new_bookmarks_file,MAX_STR_LEN);
+		bookmarks_codepage=new_bookmarks_codepage;
+		load_bookmarks(ses);
+		reinit_list_window(&bookmark_ld);
+	} else {
+		save_only:
+		safe_strncpy(bookmarks_file,new_bookmarks_file,MAX_STR_LEN);
+		bookmarks_codepage=new_bookmarks_codepage;
+		bookmark_ld.modified=1;
+		save_bookmarks(ses);
+	}
 }
 
 
@@ -747,7 +764,8 @@ static unsigned char *convert_to_entity_string(unsigned char *str)
 	unsigned char *dst, *p, *q;
 	int size;
 	
-	for (size=1,p=str;*p;size+=*p=='&'?5:*p=='<'||*p=='>'||*p=='='?4:*p=='"'?6:1,p++);
+	for (size=1,p=str;*p;size+=*p=='&'?5:*p=='<'||*p=='>'||*p=='='?4:*p=='"'?6:1,p++)
+		;
 
 	dst=mem_alloc(size);
 	
@@ -782,7 +800,7 @@ static unsigned char *convert_to_entity_string(unsigned char *str)
 }
 
 /* writes bookmarks to disk */
-void save_bookmarks(void)
+static void save_bookmarks(struct session *ses)
 {
 	struct bookmark_list *b;
 	int depth;
@@ -790,8 +808,9 @@ void save_bookmarks(void)
 	struct conv_table *ct;
 	unsigned char *data;
 	int l;
+	int err;
 
-	if (!bookmark_ld.modified||!can_write_bookmarks)return;
+	if (!bookmark_ld.modified)return;
 	ct=get_translation_table(bookmark_ld.codepage,bookmarks_codepage);
 	data=init_str();
 	l=0;
@@ -848,9 +867,16 @@ void save_bookmarks(void)
 	"</DL><P>\n"
 	"</HTML>\n"
 	);
-	write_to_config_file(bookmarks_file, data);
+	err = write_to_config_file(bookmarks_file, data);
 	mem_free(data);
-	bookmark_ld.modified=0;
+	if (!err) {
+		bookmark_ld.modified=0;
+	} else {
+		if (ses) {
+			unsigned char *f = stracpy(bookmarks_file);
+			msg_box(ses->term, getml(f, NULL), TEXT_(T_BOOKMARK_ERROR), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_UNABLE_TO_WRITE_TO_BOOKMARK_FILE), " ", f, ": ", get_err_msg(err), NULL, NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
+		}
+	}
 }
 
 void menu_bookmark_manager(struct terminal *term,void *fcp,struct session *ses)
