@@ -61,7 +61,7 @@ unsigned long aspect_native=65536; /* Like aspect, but not influenced by
  * If you change them, don't wonder if the letters start to look strange.
  * The numers in the comments denote which height the line applies for.
  */
-float fancy_constants[64]={
+static float fancy_constants[64]={
 	0,3,		/*  1 */
 	.1,3,   	/*  2 */
 	.2,3,   	/*  3 */
@@ -96,20 +96,6 @@ float fancy_constants[64]={
 	.01,1  		/* 32 */
 };
 
-/* prototypes */
-void decimate_3(unsigned short **, int, int);
-struct letter *find_stored_letter(int *, int);
-void read_stored_data(png_structp, png_bytep, png_uint_32);
-void my_png_warning(png_structp, png_const_charp);
-void my_png_error(png_structp, png_const_charp);
-struct font_cache_entry *supply_color_cache_entry (struct graphics_driver *, struct style *, struct letter *);
-void get_underline_pos(int, int *, int *);
-int compare_font_entries(void *, void *);
-void init_font_cache(int);
-void recode_font_name(unsigned char **);
-int compare_family(unsigned char *, unsigned char *);
-int fill_style_table(int *, unsigned char *);
-
 
 /* This shall be hopefully reasonably fast and portable
  * We assume ix is <65536. If not, the letters will be smaller in
@@ -138,7 +124,8 @@ static inline int compute_width (int ix, int iy, int required_height)
 	return width;
 }
 
-struct lru font_cache; /* This is a cache for screen-ready colored bitmaps
+static struct lru font_cache;
+			/* This is a cache for screen-ready colored bitmaps
                         * of lettrs and/or alpha channels for these (are the
 			* same size, only one byte per pixel and are used
 			* for letters with an image under them )
@@ -773,7 +760,7 @@ inline static void scale_gray(unsigned char *in, int ix, int iy, unsigned char *
  * x must be >=1.
  * Performs realloc onto the buffer after decimation to save memory.
  */
-void decimate_3(unsigned short **data0, int x, int y)
+static void decimate_3(unsigned short **data0, int x, int y)
 {
 	unsigned short *data=*data0;
 	unsigned short *ahead=data;
@@ -1288,11 +1275,13 @@ void apply_gamma_exponent_24_to_48_table(unsigned short *dest, unsigned char *sr
 	}
 }
 
+#if 0
 /* Input is 0-255 (8-bit). Output is 0-255 (8-bit)*/
 unsigned char apply_gamma_single_8_to_8(unsigned char input, float gamma)
 {
 	return 255*pow(((float) input)/255,gamma)+0.5;
 }
+#endif
 
 /* Input is 0-255 (8-bit). Output is 0-65535 (16-bit)*/
 /* We assume unsigned short holds at least 16 bits. */
@@ -1343,7 +1332,7 @@ extern int n_fonts; /* Number of fonts. font number 0 is system_font (it's
  * if the letter is not found. Tries all possibilities in the style table
  * before returning NULL.
  */
-struct letter *find_stored_letter(int *style_table, int letter_number)
+static struct letter *find_stored_letter(int *style_table, int letter_number)
 {
 	int first, last, half, diff, font_index, font_number;
 
@@ -1379,7 +1368,7 @@ struct letter *find_stored_letter(int *style_table, int letter_number)
 	return letter_data+font_table[0].begin;
 }
 
-void read_stored_data(png_structp png_ptr, png_bytep data, png_uint_32 length)
+static void read_stored_data(png_structp png_ptr, png_bytep data, png_uint_32 length)
 {
 	struct read_work *work;
 
@@ -1390,11 +1379,11 @@ void read_stored_data(png_structp png_ptr, png_bytep data, png_uint_32 length)
 	work->pointer+=length;
 }
 
-void my_png_warning(png_structp a, png_const_charp b)
+static void my_png_warning(png_structp a, png_const_charp b)
 {
 }
 
-void my_png_error(png_structp a, png_const_charp error_string)
+static void my_png_error(png_structp a, png_const_charp error_string)
 {
 	error("Error when loading compiled-in font: %s.",error_string);
 }
@@ -1462,7 +1451,7 @@ unsigned char *png_data, int png_length, struct style *style)
 		if (color_type==PNG_COLOR_TYPE_PALETTE){
 			png_set_expand(png_ptr);
 #ifdef HAVE_PNG_SET_RGB_TO_GRAY
-			png_set_rgb_to_gray(png_ptr,1,54.0*256,183.0*256);
+			png_set_rgb_to_gray(png_ptr, 1, -1, -1);
 #else
 			goto end;
 #endif
@@ -1473,7 +1462,7 @@ unsigned char *png_data, int png_length, struct style *style)
 		if (color_type==PNG_COLOR_TYPE_RGB ||
 			color_type==PNG_COLOR_TYPE_RGB_ALPHA){
 #ifdef HAVE_PNG_SET_RGB_TO_GRAY
-			png_set_rgb_to_gray(png_ptr, 1, 54.0*256, 183.0*256);
+			png_set_rgb_to_gray(png_ptr, 1, -1, -1);
 #else
 			goto end;
 #endif
@@ -1567,10 +1556,13 @@ unsigned char *png_data, int png_length, struct style *style)
 	mem_free(interm2);
 }
 
+static struct font_cache_entry *locked_bw_entry = NULL;
+static struct font_cache_entry *locked_color_entry = NULL;
+
 /* Adds required entry into font_cache and returns pointer to the entry.
  * We assume the entry is FC_COLOR.
  */
-struct font_cache_entry *supply_color_cache_entry (struct graphics_driver
+static struct font_cache_entry *supply_color_cache_entry (struct graphics_driver
 *gd, struct style *style, struct letter *letter)
 {
 	struct font_cache_entry *found, *new;
@@ -1594,8 +1586,11 @@ struct font_cache_entry *supply_color_cache_entry (struct graphics_driver
 			found->bitmap.y, font_data+letter->begin,
 			letter->length, style);
 		do_free=1;
+	} else {
+		locked_bw_entry = found;
 	}
 	new=mem_alloc(sizeof(*new));
+	locked_color_entry = new;
 	new->type=FC_COLOR;
 	new->bitmap=found->bitmap;
 	new->r0=style->r0;
@@ -1628,16 +1623,20 @@ struct font_cache_entry *supply_color_cache_entry (struct graphics_driver
 		decimate_3(&primary_data,new->bitmap.x,new->bitmap.y);
 	}
 	/* We have a buffer with photons */
-	gd->get_empty_bitmap(&(new->bitmap));
+	if (gd->get_empty_bitmap(&(new->bitmap)))
+		goto skip_dither;
 	if (dither_letters)
 		dither(primary_data, &(new->bitmap));
 	else
 		(*round_fn)(primary_data,&(new->bitmap));
+	skip_dither:
 	mem_free(primary_data);
 	gd->register_bitmap(&(new->bitmap));
 	if (do_free){
 		mem_free(found->bitmap.data);
 		mem_free(found);
+	} else {
+		locked_bw_entry = NULL;
 	}
 	bytes_consumed=new->bitmap.x*new->bitmap.y*(gd->depth&7);
 	/* Number of bytes per pixel in passed bitmaps */
@@ -1648,31 +1647,35 @@ struct font_cache_entry *supply_color_cache_entry (struct graphics_driver
 	return new;
 }
 
-/* Prunes the cache to comply with maximum size */
-static inline void prune_font_cache(struct graphics_driver *gd)
+static int destroy_font_cache_bottom(void)
 {
 	struct font_cache_entry *bottom;
+	bottom=lru_get_bottom(&font_cache);
+	if (!bottom) return 0;
+	if (bottom == locked_bw_entry || bottom == locked_color_entry) return 0;
+	if (bottom->type==FC_COLOR){
+		drv->unregister_bitmap(&(bottom->bitmap));
+	}else{
+		mem_free(bottom->bitmap.data);
+	}
+	mem_free(bottom);
+	lru_destroy_bottom(&font_cache);
+	return 1;
+}
 
-	while(font_cache.bytes>font_cache.max_bytes){
-		/* Prune bottom entry of font cache */
-		bottom=lru_get_bottom(&font_cache);
-		if (!bottom){
-#ifdef DEBUG
-			if (font_cache.bytes||font_cache.items){
-				internal("font cache is empty and contains\
-some items or bytes at the same time.\n");
-			}
-#endif
+/* Prunes the cache to comply with maximum size */
+static int prune_font_cache(void)
+{
+	int r = 0;
+
+	while (font_cache.bytes > (unsigned)font_cache_size) {
+		if (destroy_font_cache_bottom()) {
+			r = 1;
+		} else {
 			break;
 		}
-		if (bottom->type==FC_COLOR){
-			gd->unregister_bitmap(&(bottom->bitmap));
-		}else{
-			mem_free(bottom->bitmap.data);
-		}
-		mem_free(bottom);
-		lru_destroy_bottom(&font_cache);
 	}
+	return r;
 }
 
 /* Prints a letter to the specified position and
@@ -1682,6 +1685,7 @@ inline static int print_letter(struct graphics_driver *gd, struct
 	int char_number)
 
 {	
+	int xw;
 	struct font_cache_entry *found;
 	struct font_cache_entry template;
 	struct letter *letter;
@@ -1704,9 +1708,13 @@ inline static int print_letter(struct graphics_driver *gd, struct
 
 	found=lru_lookup(&font_cache, &template, letter->color_list);
 	if (!found) found=supply_color_cache_entry(gd, style, letter);
+	else locked_color_entry = found;
 	gd->draw_bitmap(device, &(found->bitmap), x, y);
-	prune_font_cache(gd);
-	return found->bitmap.x;
+	xw = found->bitmap.x;
+	if (locked_color_entry != found) internal("bad letter lock");
+	locked_color_entry = NULL;
+	prune_font_cache();
+	return xw;
 }
 
 /* Must return values that are:
@@ -1715,7 +1723,7 @@ inline static int print_letter(struct graphics_driver *gd, struct
  * at least 1 apart
  * Otherwise g_print_text will print nonsense (but won't segfault)
  */
-void get_underline_pos(int height, int *top, int *bottom)
+static void get_underline_pos(int height, int *top, int *bottom)
 {
 	int thickness, baseline;
 	thickness=(height+15)/16;
@@ -1814,7 +1822,7 @@ int x, int y, struct style *style, unsigned char *text, int *width)
 }
 
 /* 0=equality 1=inequality */
-int compare_font_entries(void *entry, void *template)
+static int compare_font_entries(void *entry, void *template)
 {
 	struct font_cache_entry*e1=entry;
 	struct font_cache_entry*e2=template;
@@ -1839,29 +1847,21 @@ int compare_font_entries(void *entry, void *template)
 /* If the cache already exists, it is destroyed and reallocated. If you call it with the same
  * size argument, only a cache flush will yield.
  */
-void init_font_cache(int bytes)
+static void init_font_cache(void)
 {
-	lru_init(&font_cache, &compare_font_entries, bytes);
+	lru_init(&font_cache, &compare_font_entries);
 }
 
 /* Ensures there are no lru_entry objects allocated - destroys them.
  * Also destroys the bitmaps asociated with them. Does not destruct the
  font_cache per se.
  */
-void destroy_font_cache(void)
+static void destroy_font_cache(void)
 {
-	struct font_cache_entry *bottom;
-	
-	while((bottom=lru_get_bottom(&font_cache))){
-		if (bottom->type==FC_COLOR){
-			drv->unregister_bitmap(&(bottom->bitmap));
-		}else{
-			/* Then it must be FC_BW. */
-			mem_free(bottom->bitmap.data);
-		}
-		mem_free(bottom);
-		lru_destroy_bottom(&font_cache);
-	}
+	while (destroy_font_cache_bottom())
+		;
+	if (lru_get_bottom(&font_cache))
+		internal("destroy_font_cache: cache not freed due to locks");
 }
 
 /* Returns 0 in case the char is not found. */
@@ -1927,21 +1927,49 @@ int g_wrap_text(struct wrap_struct *w)
 void update_aspect(void)
 {
 	aspect=aspect_on?(aspect_native*bfu_aspect+0.5):65536UL;
+	destroy_font_cache();
+}
+
+long fontcache_info(int type)
+{
+	switch (type) {
+		case CI_BYTES:
+			return font_cache.bytes;
+		case CI_FILES:
+			return font_cache.items;
+		default:
+			internal("fontcache_info: query %d", type);
+			return 0;
+	}
+}
+
+static int shrink_font_cache(int u)
+{
+	int freed_something = 0;
+	int has_something;
+	if (u == SH_CHECK_QUOTA) {
+		freed_something = prune_font_cache();
+	}
+	if (u == SH_FREE_ALL) {
+		while (destroy_font_cache_bottom())
+			freed_something = 1;
+	}
+	if (u == SH_FREE_SOMETHING) {
+		freed_something = destroy_font_cache_bottom();
+	}
+	has_something = !!lru_get_bottom(&font_cache);
+	return	(freed_something ? ST_SOMETHING_FREED : 0) |
+		(has_something ? 0 : ST_CACHE_EMPTY);
 }
 
 void init_dip(void)
 {
+	init_font_cache();
 	update_aspect();
-	/* Initializes to 2 MByte */
-	init_font_cache(2000000);
+	register_cache_upcall(shrink_font_cache, "fontcache");
 }
 
-void shutdown_dip(void)
-{
-	destroy_font_cache();
-}
-
-void recode_font_name(unsigned char **name)
+static void recode_font_name(unsigned char **name)
 {
 	int dashes=0;
 	unsigned char *p;
@@ -1961,7 +1989,7 @@ void recode_font_name(unsigned char **name)
  * 0 matches
  * 1 doesn't match
  */
-int compare_family(unsigned char *single, unsigned char *multi)
+static int compare_family(unsigned char *single, unsigned char *multi)
 {
 	unsigned char *p,*r;
 	int single_length=strlen(single);
@@ -1982,7 +2010,7 @@ int compare_family(unsigned char *single, unsigned char *multi)
  * will result deterministically random results).
  * Returns 1 if the font is monospaced or 0 if not.
  */
-int fill_style_table(int * table, unsigned char *name)
+static int fill_style_table(int * table, unsigned char *name)
 {
 	unsigned char *p;
 	unsigned char *family, *weight, *slant, *adstyl, *spacing;
@@ -2146,8 +2174,10 @@ void get_links_icon(unsigned char **data, int *width, int* height, int depth)
 	*width=b.x;
 	*height=b.y;
 	b.skip=b.x*(depth&7);
+	retry:
 	if (!(b.data=*data=malloc(b.skip*b.y))) {
-		malloc_oom();
+		out_of_memory("icon malloc", b.skip*b.y);
+		goto retry;
 	}
 	tmp1=mem_alloc(6*b.y*b.x);
         apply_gamma_exponent_24_to_48(tmp1,links_icon,b.x*b.y,g,g,g);
