@@ -22,29 +22,28 @@ struct ftp_connection_info {
 	unsigned char cmdbuf[1];
 };
 
-void ftp_get_banner(struct connection *);
-void ftp_got_banner(struct connection *, struct read_buffer *);
-void ftp_login(struct connection *);
-void ftp_logged(struct connection *);
-void ftp_sent_passwd(struct connection *);
-void ftp_got_reply(struct connection *, struct read_buffer *);
-void ftp_got_info(struct connection *, struct read_buffer *);
-void ftp_got_user_info(struct connection *, struct read_buffer *);
-void ftp_dummy_info(struct connection *, struct read_buffer *);
-void ftp_pass_info(struct connection *, struct read_buffer *);
-void ftp_send_retr_req(struct connection *, int);
-struct ftp_connection_info *add_file_cmd_to_str(struct connection *);
-void ftp_retr_1(struct connection *);
-void ftp_retr_file(struct connection *, struct read_buffer *);
-void ftp_got_final_response(struct connection *, struct read_buffer *);
-void created_data_connection(struct connection *);
-void got_something_from_data_connection(struct connection *);
-void ftp_end_request(struct connection *);
-int get_ftp_response(struct connection *, struct read_buffer *, int);
-int ftp_process_dirlist(struct cache_entry *, off_t *, int *, unsigned char *, int, int, int *);
+static void ftp_get_banner(struct connection *);
+static void ftp_got_banner(struct connection *, struct read_buffer *);
+static void ftp_login(struct connection *);
+static void ftp_logged(struct connection *);
+static void ftp_sent_passwd(struct connection *);
+static void ftp_got_info(struct connection *, struct read_buffer *);
+static void ftp_got_user_info(struct connection *, struct read_buffer *);
+static void ftp_dummy_info(struct connection *, struct read_buffer *);
+static void ftp_pass_info(struct connection *, struct read_buffer *);
+static void ftp_send_retr_req(struct connection *, int);
+static struct ftp_connection_info *add_file_cmd_to_str(struct connection *);
+static void ftp_retr_1(struct connection *);
+static void ftp_retr_file(struct connection *, struct read_buffer *);
+static void ftp_got_final_response(struct connection *, struct read_buffer *);
+static void created_data_connection(struct connection *);
+static void got_something_from_data_connection(struct connection *);
+static void ftp_end_request(struct connection *);
+static int get_ftp_response(struct connection *, struct read_buffer *, int);
+static int ftp_process_dirlist(struct cache_entry *, off_t *, int *, unsigned char *, int, int, int *);
 
 
-int get_ftp_response(struct connection *c, struct read_buffer *rb, int part)
+static int get_ftp_response(struct connection *c, struct read_buffer *rb, int part)
 {
 	int l;
 	set_timeout(c);
@@ -91,7 +90,7 @@ void ftp_func(struct connection *c)
 	} else ftp_send_retr_req(c, S_SENT);
 }
 
-void ftp_get_banner(struct connection *c)
+static void ftp_get_banner(struct connection *c)
 {
 	struct read_buffer *rb;
 	set_timeout(c);
@@ -100,7 +99,7 @@ void ftp_get_banner(struct connection *c)
 	read_from_socket(c, c->sock1, rb, ftp_got_banner);
 }
 
-void ftp_got_banner(struct connection *c, struct read_buffer *rb)
+static void ftp_got_banner(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
@@ -109,7 +108,7 @@ void ftp_got_banner(struct connection *c, struct read_buffer *rb)
 	ftp_login(c);
 }
 
-void ftp_login(struct connection *c)
+static void ftp_login(struct connection *c)
 {
 	unsigned char *login;
 	unsigned char *u;
@@ -138,7 +137,7 @@ void ftp_login(struct connection *c)
 	setcstate(c, S_SENT);
 }
 
-void ftp_logged(struct connection *c)
+static void ftp_logged(struct connection *c)
 {
 	struct read_buffer *rb;
 	if (!(rb = alloc_read_buffer(c))) return;
@@ -149,7 +148,7 @@ void ftp_logged(struct connection *c)
 	read_from_socket(c, c->sock1, rb, ftp_got_info);
 }
 
-void ftp_got_info(struct connection *c, struct read_buffer *rb)
+static void ftp_got_info(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
@@ -158,7 +157,7 @@ void ftp_got_info(struct connection *c, struct read_buffer *rb)
 	ftp_got_user_info(c, rb);
 }
 
-void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
+static void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
@@ -187,7 +186,7 @@ void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 	}
 }
 
-void ftp_dummy_info(struct connection *c, struct read_buffer *rb)
+static void ftp_dummy_info(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
@@ -195,14 +194,14 @@ void ftp_dummy_info(struct connection *c, struct read_buffer *rb)
 	ftp_retr_file(c, rb);
 }
 
-void ftp_sent_passwd(struct connection *c)
+static void ftp_sent_passwd(struct connection *c)
 {
 	struct read_buffer *rb;
 	if (!(rb = alloc_read_buffer(c))) return;
 	read_from_socket(c, c->sock1, rb, ftp_pass_info);
 }
 
-void ftp_pass_info(struct connection *c, struct read_buffer *rb)
+static void ftp_pass_info(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
@@ -213,7 +212,7 @@ void ftp_pass_info(struct connection *c, struct read_buffer *rb)
 	else ftp_send_retr_req(c, S_GETH);
 }
 
-struct ftp_connection_info *add_file_cmd_to_str(struct connection *c)
+static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c)
 {
 	unsigned char *d = get_url_data(c->url);
 	unsigned char *de;
@@ -318,7 +317,7 @@ struct ftp_connection_info *add_file_cmd_to_str(struct connection *c)
 }
 
 
-void ftp_send_retr_req(struct connection *c, int state)
+static void ftp_send_retr_req(struct connection *c, int state)
 {
 	struct ftp_connection_info *fi;
 	unsigned char *login;
@@ -342,14 +341,14 @@ void ftp_send_retr_req(struct connection *c, int state)
 	setcstate(c, state);
 }
 
-void ftp_retr_1(struct connection *c)
+static void ftp_retr_1(struct connection *c)
 {
 	struct read_buffer *rb;
 	if (!(rb = alloc_read_buffer(c))) return;
 	read_from_socket(c, c->sock1, rb, ftp_retr_file);
 }
 
-void ftp_retr_file(struct connection *c, struct read_buffer *rb)
+static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 {
 	int g;
 	struct ftp_connection_info *inf = c->info;
@@ -382,19 +381,8 @@ void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 								if (j != 1) goto no_pasv;
 								pc[4] = pc[0];
 								pc[5] = pc[1];
-								{
-									unsigned a;
-									struct sockaddr_in sa;
-									socklen_t nl = sizeof(sa);
-									if (getpeername(c->sock1, (struct sockaddr *)&sa, &nl)) goto no_pasv;
-									if (nl != sizeof(sa)) goto no_pasv;
-									a = ntohl(sa.sin_addr.s_addr);
-									pc[0] = a >> 24;
-									pc[1] = a >> 16;
-									pc[2] = a >> 8;
-									pc[3] = a;
-									goto pasv_ok;
-								}
+								pc[0] = pc[1] = pc[2] = pc[3] = 0;
+								goto pasv_ok;
 							}
 						}
 					}
@@ -416,41 +404,12 @@ void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 			case 2:		/* PORT */
 				if (g >= 400) { setcstate(c, S_FTP_PORT); abort_connection(c); return; }
 				if (inf->pasv) {
-#if 0
-					struct sockaddr_in sa;
-					if (!pc[0] && !pc[1] && !pc[2] && !pc[3] && !pc[4] && !pc[5]) {
+					if (!pc[4] && !pc[5]) {
 						setcstate(c, S_FTP_ERROR);
 						retry_connection(c);
 						return;
 					}
-					sa.sin_family = AF_INET;
-					sa.sin_port = htons((pc[4] << 8) + pc[5]);
-					sa.sin_addr.s_addr = htonl((pc[0] << 24) + (pc[1] << 16) + (pc[2] << 8) + pc[3]);
-					/*debug("%d.%d.%d.%d", pc[0], pc[1], pc[2], pc[3]);*/
-					if ((c->sock2 = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
-						setcstate(c, -errno);
-						retry_connection(c);
-						return;
-					}
-					fcntl(c->sock2, F_SETFL, O_NONBLOCK);
-#ifdef HAVE_IPTOS
-					if (ftp_options.set_tos) {
-						int on = IPTOS_THROUGHPUT;
-						setsockopt(c->sock2, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int));
-					}
-#endif
-					if (connect(c->sock2, (struct sockaddr *)&sa, sizeof sa)) {
-						if (errno != EINPROGRESS && errno != EALREADY) {
-							setcstate(c, -errno);
-							retry_connection(c);
-							return;
-						}
-					}
-					inf->d = 1;
-					set_handlers(c->sock2, (void (*)(void *))got_something_from_data_connection, NULL, NULL, c);
-#else
 					make_connection(c, (pc[4] << 8) + pc[5], &c->sock2, created_data_connection);
-#endif
 				}
 				goto rep;
 			case 3:		/* REST / CWD */
@@ -497,7 +456,7 @@ void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 	ftp_got_final_response(c, rb);
 }
 
-void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
+static void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 {
 	struct ftp_connection_info *inf = c->info;
 	int g = get_ftp_response(c, rb, 0);
@@ -505,10 +464,13 @@ void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 	if (!g) { read_from_socket(c, c->sock1, rb, ftp_got_final_response); if (c->state != S_TRANS) setcstate(c, S_GETH); return; }
 	if (g == 425 || g == 450 || g == 500 || g == 501 || g == 550) {
 		if (c->url[strlen(c->url) - 1] == '/') goto skip_redir;
-		if (!c->cache && get_cache_entry(c->url, &c->cache)) {
-			setcstate(c, S_OUT_OF_MEM);
-			abort_connection(c);
-			return;
+		if (!c->cache) {
+			if (get_cache_entry(c->url, &c->cache)) {
+				setcstate(c, S_OUT_OF_MEM);
+				abort_connection(c);
+				return;
+			}
+			c->cache->refcount--;
 		}
 		if (c->cache->redirect) mem_free(c->cache->redirect);
 		c->cache->redirect = stracpy(c->url);
@@ -544,7 +506,7 @@ static int is_date(char *data)	/* can touch at most data[-4] --- "n 12 "<--if fe
 	return 1;
 }
 
-int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned char *bf, int ln, int fin, int *tr)
+static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned char *bf, int ln, int fin, int *tr)
 {
 	unsigned char *str, *buf;
 	int sl;
@@ -604,6 +566,7 @@ int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned cha
 		if (p > 5 && !casecmp(buf, "total", 5)) goto raw;
 		for (pp = p - 1; pp >= 0; pp--) if (!WHITECHAR(buf[pp])) break;
 		if (pp < 0) goto raw;
+		if (pp < p - 1) pp++;
 		ppos = -1;
 		for (; pp >= 10; pp--) if (WHITECHAR(buf[pp])) {
 			if (is_date(&buf[pp - 6]) &&
@@ -612,7 +575,7 @@ int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned cha
 			     (buf[pp - 4] == '1' && buf[pp - 3] == '9')) &&
 			    buf[pp - 2] >= '0' && buf[pp - 2] <= '9' &&
 			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') {
-				if (pp < p - 1 && buf[pp + 1] == ' ') ppos = pp + 1;
+				if (pp < p - 2 && buf[pp + 1] == ' ' && buf[pp + 2] != ' ') ppos = pp + 1;
 				else ppos = pp;
 			}
 			if (buf[pp - 6] == ' ' &&
@@ -620,7 +583,11 @@ int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned cha
 			    buf[pp - 4] >= '0' && buf[pp - 4] <= '9' &&
 			    buf[pp - 3] == ':' &&
 			    buf[pp - 2] >= '0' && buf[pp - 2] <= '5' &&
-			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') ppos = pp;
+			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') {
+				ppos = pp;
+				if (pp + 2 < p && buf[pp + 1] == ' ' && buf[pp + 2] != ' ')
+					ppos++;
+			}
 		}
 		if (ppos != -1) {
 			pp = ppos;
@@ -674,7 +641,7 @@ int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned cha
 	goto again;
 }
 
-void created_data_connection(struct connection *c)
+static void created_data_connection(struct connection *c)
 {
 	struct ftp_connection_info *inf = c->info;
 #ifdef HAVE_IPTOS
@@ -687,7 +654,7 @@ void created_data_connection(struct connection *c)
 	set_handlers(c->sock2, (void (*)(void *))got_something_from_data_connection, NULL, NULL, c);
 }
 
-void got_something_from_data_connection(struct connection *c)
+static void got_something_from_data_connection(struct connection *c)
 {
 	struct ftp_connection_info *inf = c->info;
 	int l;
@@ -702,10 +669,13 @@ void got_something_from_data_connection(struct connection *c)
 		set_handlers(ns, (void (*)(void *))got_something_from_data_connection, NULL, NULL, c);
 		return;
 	}
-	if (!c->cache && get_cache_entry(c->url, &c->cache)) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
+	if (!c->cache) {
+		if (get_cache_entry(c->url, &c->cache)) {
+			setcstate(c, S_OUT_OF_MEM);
+			abort_connection(c);
+			return;
+		}
+		c->cache->refcount--;
 	}
 	if (inf->dir && !c->from) {
 		unsigned char *ud;
@@ -739,13 +709,13 @@ void got_something_from_data_connection(struct connection *c)
 			inf->conn_st = 2;
 			return;
 		}
-		setcstate(c, -errno);
+		setcstate(c, get_error_from_errno(errno));
 		retry_connection(c);
 		return;
 	}
 	if (l > 0) {
 		if (!inf->dir) {
-			if (c->from + l < 0) {
+			if ((off_t)(0UL + c->from + l) < 0) {
 				setcstate(c, S_LARGE_FILE);
 				abort_connection(c);
 				return;
@@ -774,7 +744,7 @@ void got_something_from_data_connection(struct connection *c)
 	}
 }
 
-void ftp_end_request(struct connection *c)
+static void ftp_end_request(struct connection *c)
 {
 	if (c->state == S_OK) {
 		if (c->cache) {
