@@ -16,6 +16,12 @@
 
 #define XBM_BUFFER_LEN 1024
 
+#ifdef min
+#undef min
+#endif
+#ifdef max
+#undef max
+#endif
 #define min(a,b) (a<b?a:b)
 #define max(a,b) (a>b?a:b)
 
@@ -41,13 +47,8 @@ struct xbm_decoder{
 	unsigned char buffer[XBM_BUFFER_LEN];
 };
 
-extern int get_foreground(int rgb);
 
-unsigned char *my_memmem(unsigned char *, int, unsigned char *, int);
-void xbm_decode(struct cached_image *, unsigned char *, int);
-
-
-unsigned char *my_memmem(unsigned char *h, int hl, unsigned char *n, int nl)
+static unsigned char *my_memmem(unsigned char *h, int hl, unsigned char *n, int nl)
 {
 	for (;hl>=nl;hl--,h++)
 		if (*h==*n&&!memcmp(h,n,nl))return h;
@@ -110,13 +111,15 @@ static inline int __read_num(unsigned char **p,int *l,int *partnum,int *digits,
 
 static inline void __skip_space_tab(unsigned char **p, int *l)
 {
-	for (;*l&&(**p==' '||**p==9);(*l)--,(*p)++);
+	for (;*l&&(**p==' '||**p==9);(*l)--,(*p)++)
+		;
 }
 
 
 static inline void __skip_whitespace(unsigned char **p, int *l)
 {
-	for (;*l&&((**p)>'9'||(**p)<'0');(*l)--,(*p)++);
+	for (;*l&&((**p)>'9'||(**p)<'0');(*l)--,(*p)++)
+		;
 }
 
 
@@ -138,7 +141,7 @@ static inline void put_eight(struct cached_image *cimg,int bits)
 
 /* opravdovy dekoder xbm, data jsou bez komentaru */
 /* length is always !=NULL */
-void xbm_decode(struct cached_image *cimg, unsigned char *data, int length)
+static void xbm_decode(struct cached_image *cimg, unsigned char *data, int length)
 {
 	struct xbm_decoder *deco=(struct xbm_decoder *)cimg->decoder;
 	/* okurky v decu ;-) */
@@ -220,7 +223,7 @@ restart_again:
 			cimg->green_gamma=display_green_gamma;
 			cimg->blue_gamma=display_blue_gamma;
 			cimg->strip_optimized=0;
-			header_dimensions_known(cimg);
+			if (header_dimensions_known(cimg)) {img_end(cimg);return;}
 			
 			deco->in_data_block=1;
 			p++;
